@@ -1,4 +1,4 @@
-# Copyright (c) 2017-2021 by Ron Frederick <ronf@timeheart.net> and others.
+# Copyright (c) 2017-2023 by Ron Frederick <ronf@timeheart.net> and others.
 #
 # This program and the accompanying materials are made available under
 # the terms of the Eclipse Public License v2.0 which accompanies this
@@ -23,6 +23,8 @@
 # Some of the imports below won't be found when running pylint on UNIX
 # pylint: disable=import-error
 
+from typing import Optional, Sequence, Union
+
 from sspi import ClientAuth, ServerAuth
 from sspi import error as SSPIError
 
@@ -33,74 +35,89 @@ from sspicon import ASC_RET_INTEGRITY, ASC_RET_MUTUAL_AUTH
 from sspicon import SECPKG_ATTR_NATIVE_NAMES
 
 from .asn1 import ObjectIdentifier, der_encode
+from .misc import BytesOrStrDict
 
 
 _krb5_oid = der_encode(ObjectIdentifier('1.2.840.113554.1.2.2'))
 
 
-class _GSSBase:
+class GSSBase:
     """GSS base class"""
 
     # Overridden in child classes
     _mutual_auth_flag = 0
     _integrity_flag = 0
 
-    def __init__(self, host):
+    def __init__(self, host: str):
         if '@' in host:
             self._host = host
         else:
             self._host = 'host/' + host
 
-        self._ctx = None
-        self._init_token = None
+        self._ctx: Optional[Union[ClientAuth, ServerAuth]] = None
+        self._init_token: Optional[bytes] = None
 
     @property
-    def mechs(self):
+    def mechs(self) -> Sequence[bytes]:
         """Return GSS mechanisms available for this host"""
 
         return [_krb5_oid]
 
     @property
-    def complete(self):
+    def complete(self) -> bool:
         """Return whether or not GSS negotiation is complete"""
 
+        assert self._ctx is not None
+
         return self._ctx.authenticated
 
     @property
-    def provides_mutual_auth(self):
+    def provides_mutual_auth(self) -> bool:
         """Return whether or not this context provides mutual authentication"""
 
-        return self._ctx.ctxt_attr & self._mutual_auth_flag
+        assert self._ctx is not None
+
+        return bool(self._ctx.ctxt_attr & self._mutual_auth_flag)
 
     @property
-    def provides_integrity(self):
+    def provides_integrity(self) -> bool:
         """Return whether or not this context provides integrity protection"""
 
-        return self._ctx.ctxt_attr & self._integrity_flag
+        assert self._ctx is not None
+
+        return bool(self._ctx.ctxt_attr & self._integrity_flag)
 
     @property
-    def user(self):
+    def user(self) -> str:
         """Return user principal associated with this context"""
 
+        assert self._ctx is not None
+
         names = self._ctx.ctxt.QueryContextAttributes(SECPKG_ATTR_NATIVE_NAMES)
         return names[0]
 
     @property
-    def host(self):
+    def host(self) -> str:
         """Return host principal associated with this context"""
 
+        assert self._ctx is not None
+
         names = self._ctx.ctxt.QueryContextAttributes(SECPKG_ATTR_NATIVE_NAMES)
         return names[1]
 
-    def reset(self):
+    def reset(self) -> None:
         """Reset GSS security context"""
 
-        if self._ctx.authenticated:
-            self._ctx.reset()
+        assert self._ctx is not None
 
-    def step(self, token=None):
+        self._ctx.reset()
+        self._init_token = None
+
+    def step(self, token: Optional[bytes] = None) -> Optional[bytes]:
         """Perform next step in GSS security exchange"""
 
+        assert self._ctx is not None
+
         if self._init_token:
             token = self._init_token
             self._init_token = None
@@ -112,17 +129,21 @@ class _GSSBase:
         except SSPIError as exc:
             raise GSSError(details=exc.strerror) from None
 
-    def sign(self, data):
+    def sign(self, data: bytes) -> bytes:
         """Sign a block of data"""
 
+        assert self._ctx is not None
+
         try:
             return self._ctx.sign(data)
         except SSPIError as exc:
             raise GSSError(details=exc.strerror) from None
 
-    def verify(self, data, sig):
+    def verify(self, data: bytes, sig: bytes) -> bool:
         """Verify a signature for a block of data"""
 
+        assert self._ctx is not None
+
         try:
             self._ctx.verify(data, sig)
             return True
@@ -130,13 +151,17 @@ class _GSSBase:
             return False
 
 
-class GSSClient(_GSSBase):
+class GSSClient(GSSBase):
     """GSS client"""
 
     _mutual_auth_flag = ISC_RET_MUTUAL_AUTH
     _integrity_flag = ISC_RET_INTEGRITY
 
-    def __init__(self, host, delegate_creds):
+    def __init__(self, host: str, store: Optional[BytesOrStrDict],
+                 delegate_creds: bool):
+        if store is not None:
+            raise GSSError(details='GSS store not supported on Windows')
+
         super().__init__(host)
 
         flags = ISC_REQ_MUTUAL_AUTH | ISC_REQ_INTEGRITY
@@ -153,13 +178,16 @@ class GSSClient(_GSSBase):
         self._init_token = self.step(None)
 
 
-class GSSServer(_GSSBase):
+class GSSServer(GSSBase):
     """GSS server"""
 
     _mutual_auth_flag = ASC_RET_MUTUAL_AUTH
     _integrity_flag = ASC_RET_INTEGRITY
 
-    def __init__(self, host):
+    def __init__(self, host: str, store: Optional[BytesOrStrDict]):
+        if store is not None:
+            raise GSSError(details='GSS store not supported on Windows')
+
         super().__init__(host)
 
         flags = ASC_REQ_MUTUAL_AUTH | ASC_REQ_INTEGRITY
@@ -171,9 +199,10 @@ class GSSServer(_GSSBase):
 
 
 class GSSError(Exception):
-    """Stub class for reporting that GSS is not available"""
+    """Class for reporting GSS errors"""
 
-    def __init__(self, maj_code=0, min_code=0, token=None, details=''):
+    def __init__(self, maj_code: int = 0, min_code: int = 0,
+                 token: Optional[bytes] = None, details: str = ''):
         super().__init__(details)
 
         self.maj_code = maj_code
