@@ -1,4 +1,4 @@
-# Copyright (c) 2016-2021 by Ron Frederick <ronf@timeheart.net> and others.
+# Copyright (c) 2016-2023 by Ron Frederick <ronf@timeheart.net> and others.
 #
 # This program and the accompanying materials are made available under
 # the terms of the Eclipse Public License v2.0 which accompanies this
@@ -22,17 +22,28 @@
 
 import asyncio
 import os
+from pathlib import Path
 import socket
 import time
-
-from collections import namedtuple
-from pathlib import Path
+from typing import TYPE_CHECKING, Callable, Dict, Iterable
+from typing import NamedTuple, Optional, Sequence, Set, Tuple
 
 from .constants import OPEN_CONNECT_FAILED
-from .forward import SSHForwarder
-from .listener import create_tcp_forward_listener
+from .forward import SSHForwarder, SSHForwarderCoro
+from .listener import SSHListener, create_tcp_forward_listener
 from .logging import logger
 from .misc import ChannelOpenError
+from .session import DataType
+
+
+if TYPE_CHECKING:
+    # pylint: disable=cyclic-import
+    from .channel import SSHChannel
+    from .connection import SSHServerConnection
+
+
+_RecvHandler = Optional[Callable[[bytes], None]]
+
 
 # Xauth address families
 XAUTH_FAMILY_IPV4     = 0
@@ -60,7 +71,7 @@ X11_MAX_DISPLAYS      = 64
 X11_LISTEN_HOST       = 'localhost'
 
 
-def _parse_display(display):
+def _parse_display(display: str) -> Tuple[str, str, int]:
     """Parse an X11 display value"""
 
     try:
@@ -80,7 +91,8 @@ def _parse_display(display):
 
     return host, dpynum, screen
 
-async def _lookup_host(loop, host, family):
+async def _lookup_host(loop: asyncio.AbstractEventLoop, host: str,
+                       family: int) -> Sequence[str]:
     """Look up IPv4 or IPv6 addresses of a host name"""
 
     try:
@@ -92,19 +104,24 @@ async def _lookup_host(loop, host, family):
     return [ai[4][0] for ai in addrinfo]
 
 
-class SSHXAuthorityEntry(namedtuple('SSHXAuthorityEntry',
-                                    'family addr dpynum proto data')):
+class SSHXAuthorityEntry(NamedTuple):
     """An entry in an Xauthority file"""
 
-    def __bytes__(self):
+    family: int
+    addr: bytes
+    dpynum: bytes
+    proto: bytes
+    data: bytes
+
+    def __bytes__(self) -> bytes:
         """Construct an Xauthority entry"""
 
-        def _uint16(value):
+        def _uint16(value: int) -> bytes:
             """Construct a big-endian 16-bit unsigned integer"""
 
             return value.to_bytes(2, 'big')
 
-        def _string(data):
+        def _string(data: bytes) -> bytes:
             """Construct a binary string with a 16-bit length"""
 
             return _uint16(len(data)) + data
@@ -117,14 +134,14 @@ class SSHXAuthorityEntry(namedtuple('SSHXAuthorityEntry',
 class SSHX11ClientForwarder(SSHForwarder):
     """X11 forwarding connection handler"""
 
-    def __init__(self, listener, peer):
+    def __init__(self, listener: 'SSHX11ClientListener', peer: SSHForwarder):
         super().__init__(peer)
 
         self._listener = listener
 
         self._inpbuf = b''
         self._bytes_needed = 12
-        self._recv_handler = self._recv_prefix
+        self._recv_handler: _RecvHandler = self._recv_prefix
 
         self._endian = b''
         self._prefix = b''
@@ -137,7 +154,7 @@ class SSHX11ClientForwarder(SSHForwarder):
         self._auth_data = b''
         self._auth_data_pad = b''
 
-    def _encode_uint16(self, value):
+    def _encode_uint16(self, value: int) -> bytes:
         """Encode a 16-bit unsigned integer"""
 
         if self._endian == b'B':
@@ -145,7 +162,7 @@ class SSHX11ClientForwarder(SSHForwarder):
         else:
             return bytes((value & 255, value >> 8))
 
-    def _decode_uint16(self, value):
+    def _decode_uint16(self, value: bytes) -> int:
         """Decode a 16-bit unsigned integer"""
 
         if self._endian == b'B':
@@ -154,19 +171,19 @@ class SSHX11ClientForwarder(SSHForwarder):
             return (value[1] << 8) + value[0]
 
     @staticmethod
-    def _padded_len(length):
+    def _padded_len(length: int) -> int:
         """Return length rounded up to the next multiple of 4 bytes"""
 
         return ((length + 3) // 4) * 4
 
     @staticmethod
-    def _pad(data):
+    def _pad(data: bytes) -> bytes:
         """Pad a string to a multiple of 4 bytes"""
 
         length = len(data) % 4
         return data + ((4 - length) * b'\00' if length else b'')
 
-    def _recv_prefix(self, data):
+    def _recv_prefix(self, data: bytes) -> None:
         """Parse X11 client prefix"""
 
         self._endian = data[:1]
@@ -178,7 +195,7 @@ class SSHX11ClientForwarder(SSHForwarder):
         self._recv_handler = self._recv_auth_proto
         self._bytes_needed = self._padded_len(self._auth_proto_len)
 
-    def _recv_auth_proto(self, data):
+    def _recv_auth_proto(self, data: bytes) -> None:
         """Extract X11 auth protocol"""
 
         self._auth_proto = data[:self._auth_proto_len]
@@ -187,7 +204,7 @@ class SSHX11ClientForwarder(SSHForwarder):
         self._recv_handler = self._recv_auth_data
         self._bytes_needed = self._padded_len(self._auth_data_len)
 
-    def _recv_auth_data(self, data):
+    def _recv_auth_data(self, data: bytes) -> None:
         """Extract X11 auth data"""
 
         self._auth_data = data[:self._auth_data_len]
@@ -219,13 +236,13 @@ class SSHX11ClientForwarder(SSHForwarder):
         self._recv_handler = None
         self._bytes_needed = 0
 
-    def data_received(self, data, datatype=None):
+    def data_received(self, data: bytes, datatype: DataType = None) -> None:
         """Handle incoming data from the X11 client"""
 
         if self._recv_handler:
             self._inpbuf += data
 
-            while self._recv_handler:
+            while self._recv_handler: # type: ignore[truthy-function]
                 if len(self._inpbuf) >= self._bytes_needed:
                     data = self._inpbuf[:self._bytes_needed]
                     self._inpbuf = self._inpbuf[self._bytes_needed:]
@@ -243,15 +260,16 @@ class SSHX11ClientForwarder(SSHForwarder):
 class SSHX11ClientListener:
     """Client listener used to accept forwarded X11 connections"""
 
-    def __init__(self, loop, host, dpynum, auth_proto, auth_data):
+    def __init__(self, loop: asyncio.AbstractEventLoop, host: str, dpynum: str,
+                 auth_proto: bytes, auth_data: bytes):
         self._host = host
         self._dpynum = dpynum
         self._auth_proto = auth_proto
         self._local_auth = auth_data
 
         if host.startswith('/'):
-            self._connect_coro = loop.create_unix_connection
-            self._connect_args = (host + ':' + dpynum,)
+            self._connect_coro: SSHForwarderCoro = loop.create_unix_connection
+            self._connect_args: Sequence[object] = (host + ':' + dpynum,)
         elif host in ('', 'unix'):
             self._connect_coro = loop.create_unix_connection
             self._connect_args = ('/tmp/.X11-unix/X' + dpynum,)
@@ -259,10 +277,11 @@ class SSHX11ClientListener:
             self._connect_coro = loop.create_connection
             self._connect_args = (host, X11_BASE_PORT + int(dpynum))
 
-        self._remote_auth = {}
-        self._channel = {}
+        self._remote_auth: Dict['SSHChannel', bytes] = {}
+        self._channel: Dict[bytes, Tuple['SSHChannel', bool]] = {}
 
-    def attach(self, display, chan, single_connection):
+    def attach(self, display: str, chan: 'SSHChannel',
+               single_connection: bool) -> Tuple[bytes, bytes, int]:
         """Attach a channel to this listener"""
 
         host, dpynum, screen = _parse_display(display)
@@ -277,7 +296,7 @@ class SSHX11ClientListener:
 
         return self._auth_proto, remote_auth, screen
 
-    def detach(self, chan):
+    def detach(self, chan: 'SSHChannel') -> bool:
         """Detach a channel from this listener"""
 
         try:
@@ -286,11 +305,13 @@ class SSHX11ClientListener:
         except KeyError:
             pass
 
-        return self._remote_auth == {}
+        return not bool(self._remote_auth)
 
-    async def forward_connection(self):
+    async def forward_connection(self) -> SSHX11ClientForwarder:
         """Forward an incoming connection to the local X server"""
 
+        peer: SSHForwarder
+
         try:
             _, peer = await self._connect_coro(SSHForwarder,
                                                *self._connect_args)
@@ -299,7 +320,7 @@ class SSHX11ClientListener:
 
         return SSHX11ClientForwarder(self, peer)
 
-    def validate_auth(self, remote_auth):
+    def validate_auth(self, remote_auth: bytes) -> bytes:
         """Validate client auth and enforce single connection flag"""
 
         chan, single_connection = self._channel[remote_auth]
@@ -314,19 +335,19 @@ class SSHX11ClientListener:
 class SSHX11ServerListener:
     """Server listener used to forward X11 connections"""
 
-    def __init__(self, tcp_listener, display):
+    def __init__(self, tcp_listener: SSHListener, display: str):
         self._tcp_listener = tcp_listener
         self._display = display
-        self._channels = set()
+        self._channels: Set[object] = set()
 
-    def attach(self, chan, screen):
+    def attach(self, chan: 'SSHChannel', screen: int) -> str:
         """Attach a channel to this listener and return its display"""
 
         self._channels.add(chan)
 
         return '%s.%s' % (self._display, screen)
 
-    def detach(self, chan):
+    def detach(self, chan: 'SSHChannel') -> bool:
         """Detach a channel from this listener"""
 
         try:
@@ -336,13 +357,12 @@ class SSHX11ServerListener:
 
         if not self._channels:
             self._tcp_listener.close()
-            self._tcp_listener = None
             return True
         else:
             return False
 
 
-def get_xauth_path(auth_path):
+def get_xauth_path(auth_path: Optional[str]) -> str:
     """Compute the path to the Xauthority file"""
 
     if not auth_path:
@@ -354,10 +374,10 @@ def get_xauth_path(auth_path):
     return auth_path
 
 
-def walk_xauth(auth_path):
+def walk_xauth(auth_path: str) -> Iterable[SSHXAuthorityEntry]:
     """Walk the entries in an Xauthority file"""
 
-    def _read_bytes(n):
+    def _read_bytes(n: int) -> bytes:
         """Read exactly n bytes"""
 
         data = auth_file.read(n)
@@ -367,12 +387,12 @@ def walk_xauth(auth_path):
 
         return data
 
-    def _read_uint16():
+    def _read_uint16() -> int:
         """Read a 16-bit unsigned integer"""
 
         return int.from_bytes(_read_bytes(2), 'big')
 
-    def _read_string():
+    def _read_string() -> bytes:
         """Read a string"""
 
         return _read_bytes(_read_uint16())
@@ -395,7 +415,9 @@ def walk_xauth(auth_path):
         pass
 
 
-async def lookup_xauth(loop, auth_path, host, dpynum):
+async def lookup_xauth(loop: asyncio.AbstractEventLoop,
+                       auth_path: Optional[str], host: str,
+                       dpynum: str) -> Tuple[bytes, bytes]:
     """Look up Xauthority data for the specified display"""
 
     auth_path = get_xauth_path(auth_path)
@@ -405,8 +427,8 @@ async def lookup_xauth(loop, auth_path, host, dpynum):
 
     dpynum = dpynum.encode('ascii')
 
-    ipv4_addrs = []
-    ipv6_addrs = []
+    ipv4_addrs: Sequence[str] = []
+    ipv6_addrs: Sequence[str] = []
 
     for entry in walk_xauth(auth_path):
         if entry.dpynum and entry.dpynum != dpynum:
@@ -437,7 +459,9 @@ async def lookup_xauth(loop, auth_path, host, dpynum):
     logger.debug1('No xauth entry found for display: using random auth')
     return XAUTH_PROTO_COOKIE, os.urandom(XAUTH_COOKIE_LEN)
 
-async def update_xauth(auth_path, host, dpynum, auth_proto, auth_data):
+
+async def update_xauth(auth_path: Optional[str], host: str, dpynum: str,
+                       auth_proto: bytes, auth_data: bytes) -> None:
     """Update Xauthority data for the specified display"""
 
     if host.startswith('/') or host in ('', 'unix', 'localhost'):
@@ -467,8 +491,8 @@ async def update_xauth(auth_path, host, dpynum, auth_proto, auth_data):
     if not new_file:
         raise ValueError('Unable to acquire Xauthority lock')
 
-    new_entry = SSHXAuthorityEntry(XAUTH_FAMILY_HOSTNAME, host, dpynum,
-                                   auth_proto, auth_data)
+    new_entry = SSHXAuthorityEntry(XAUTH_FAMILY_HOSTNAME, host,
+                                   dpynum, auth_proto, auth_data)
 
     new_file.write(bytes(new_entry))
 
@@ -482,7 +506,10 @@ async def update_xauth(auth_path, host, dpynum, auth_proto, auth_data):
     os.replace(new_auth_path, auth_path)
 
 
-async def create_x11_client_listener(loop, display, auth_path):
+async def create_x11_client_listener(loop: asyncio.AbstractEventLoop,
+                                     display: str,
+                                     auth_path: Optional[str]) -> \
+        SSHX11ClientListener:
     """Create a listener to accept X11 connections forwarded over SSH"""
 
     host, dpynum, _ = _parse_display(display)
@@ -492,8 +519,11 @@ async def create_x11_client_listener(loop, display, auth_path):
     return SSHX11ClientListener(loop, host, dpynum, auth_proto, auth_data)
 
 
-async def create_x11_server_listener(conn, loop, auth_path,
-                                     auth_proto, auth_data):
+async def create_x11_server_listener(conn: 'SSHServerConnection',
+                                     loop: asyncio.AbstractEventLoop,
+                                     auth_path: Optional[str],
+                                     auth_proto: bytes, auth_data: bytes) -> \
+        Optional[SSHX11ServerListener]:
     """Create a listener to forward X11 connections over SSH"""
 
     for dpynum in range(X11_DISPLAY_START, X11_MAX_DISPLAYS):
@@ -507,7 +537,7 @@ async def create_x11_server_listener(conn, loop, auth_path,
         display = '%s:%d' % (X11_LISTEN_HOST, dpynum)
 
         try:
-            await update_xauth(auth_path, X11_LISTEN_HOST, dpynum,
+            await update_xauth(auth_path, X11_LISTEN_HOST, str(dpynum),
                                auth_proto, auth_data)
         except ValueError:
             tcp_listener.close()
