@@ -96,7 +96,7 @@ class _AuthClientStub(_AuthConnectionStub):
                                          password_change_prompt, kbdint_auth,
                                          kbdint_challenge, success), False)
 
-        self._gss = GSSClient(gss_host, False) if gss_host else None
+        self._gss = GSSClient(gss_host, None, False) if gss_host else None
 
         self._client_host_key = client_host_key
         self._client_host_cert = client_host_cert
@@ -259,7 +259,7 @@ class _AuthServerStub(_AuthConnectionStub):
                  kbdint_challenge=False, success=False):
         super().__init__(peer, True)
 
-        self._gss = GSSServer(gss_host) if gss_host else None
+        self._gss = GSSServer(gss_host, None) if gss_host else None
         self._override_gss_mech = override_gss_mech
 
         self._host_based_auth = host_based_auth
@@ -517,7 +517,7 @@ class _TestAuth(AsyncTestCase):
     async def test_hostbased_auth(self):
         """Unit test host-based authentication"""
 
-        hkey = get_test_key('ssh-rsa')
+        hkey = get_test_key('ecdsa-sha2-nistp256')
         cert = hkey.generate_host_certificate(hkey, 'host')
 
         with self.subTest('Host-based auth not available'):
@@ -541,7 +541,7 @@ class _TestAuth(AsyncTestCase):
     async def test_publickey_auth(self):
         """Unit test public key authentication"""
 
-        ckey = get_test_key('ssh-rsa')
+        ckey = get_test_key('ecdsa-sha2-nistp256')
         cert = ckey.generate_user_certificate(ckey, 'name')
 
         with self.subTest('Public key auth not available'):
