@@ -1,4 +1,4 @@
-# Copyright (c) 2016-2020 by Ron Frederick <ronf@timeheart.net> and others.
+# Copyright (c) 2016-2024 by Ron Frederick <ronf@timeheart.net> and others.
 #
 # This program and the accompanying materials are made available under
 # the terms of the Eclipse Public License v2.0 which accompanies this
@@ -40,6 +40,8 @@ from asyncssh.constants import MSG_CHANNEL_SUCCESS
 from asyncssh.packet import Byte, String, UInt32
 from asyncssh.public_key import CERT_TYPE_USER
 from asyncssh.stream import SSHTCPStreamSession, SSHUNIXStreamSession
+from asyncssh.stream import SSHTunTapStreamSession
+from asyncssh.tuntap import SSH_TUN_MODE_POINTTOPOINT, SSH_TUN_MODE_ETHERNET
 
 from .server import Server, ServerTestCase
 from .util import asynctest, echo, make_certificate
@@ -63,6 +65,11 @@ class _ClientChannel(asyncssh.SSHClientChannel):
 
         super()._send_request(request, *args, want_reply=want_reply)
 
+    def get_send_pktsize(self):
+        """Return the sender's max packet size """
+
+        return self._send_pktsize
+
     def send_request(self, request, *args):
         """Send a custom request (for unit testing)"""
 
@@ -151,6 +158,11 @@ class _ServerChannel(asyncssh.SSHServerChannel):
 
         asyncio.get_event_loop().call_later(0.1, self._report_response, True)
 
+    def get_send_pktsize(self):
+        """Return the sender's max packet size """
+
+        return self._send_pktsize
+
     async def open_session(self):
         """Attempt to open a session on the client"""
 
@@ -158,7 +170,7 @@ class _ServerChannel(asyncssh.SSHServerChannel):
 
 
 class _EchoServerSession(asyncssh.SSHServerSession):
-    """A shell session which echos data from stdin to stdout/stderr"""
+    """A shell session which echoes data from stdin to stdout/stderr"""
 
     def __init__(self):
         self._chan = None
@@ -283,8 +295,9 @@ class _ChannelServer(Server):
             except asyncssh.ChannelOpenError:
                 stdout.channel.exit(1)
         elif action == 'rejected_session':
-            chan = _ServerChannel(self._conn, asyncio.get_event_loop(), False,
-                                  False, 0, 1024, None, 'strict', 1, 32768)
+            chan = _ServerChannel(self._conn, asyncio.get_event_loop(),
+                                  False, False, False, 0, 1024, None,
+                                  'strict', 1, 32768)
 
             try:
                 await chan.open_session()
@@ -332,6 +345,22 @@ class _ChannelServer(Server):
                 await chan.accept(SSHUNIXStreamSession, b'\xff')
             except asyncssh.ChannelOpenError:
                 stdout.channel.exit(1)
+        elif action == 'rejected_tun_request':
+            chan = self._conn.create_tuntap_channel()
+
+            try:
+                await chan.open(SSHTunTapStreamSession,
+                                SSH_TUN_MODE_POINTTOPOINT, 0)
+            except asyncssh.ChannelOpenError:
+                stdout.channel.exit(1)
+        elif action == 'rejected_tap_request':
+            chan = self._conn.create_tuntap_channel()
+
+            try:
+                await chan.open(SSHTunTapStreamSession,
+                                SSH_TUN_MODE_ETHERNET, 0)
+            except asyncssh.ChannelOpenError:
+                stdout.channel.exit(1)
         elif action == 'late_auth_banner':
             try:
                 self._conn.send_auth_banner('auth banner')
@@ -398,6 +427,10 @@ class _ChannelServer(Server):
         elif action == 'data_past_window':
             stdin.channel.send_packet(MSG_CHANNEL_DATA,
                                       String(2*1025*1024*'\0'))
+        elif action == 'ext_data_past_window':
+            stdin.channel.send_packet(MSG_CHANNEL_EXTENDED_DATA,
+                                      UInt32(asyncssh.EXTENDED_DATA_STDERR),
+                                      String(2*1025*1024*'\0'))
         elif action == 'data_after_eof':
             stdin.channel.send_packet(MSG_CHANNEL_EOF)
             stdout.write('xxx')
@@ -426,6 +459,9 @@ class _ChannelServer(Server):
                                    String('ssh-connection'), String('none'))
         elif action == 'invalid_response':
             stdin.channel.send_packet(MSG_CHANNEL_SUCCESS)
+        elif action == 'send_pktsize':
+            stdout.write(str(stdout.channel.get_send_pktsize()))
+            stdout.close()
         else:
             stdin.channel.exit(255)
 
@@ -691,7 +727,7 @@ class _TestChannel(ServerTestCase):
                 chan, _ = await _create_session(conn)
 
                 result = await chan.make_request(b'keepalive@openssh.com')
-                self.assertTrue(result)
+                self.assertFalse(result)
 
     @asynctest
     async def test_invalid_open_confirmation(self):
@@ -894,6 +930,18 @@ class _TestChannel(ServerTestCase):
 
         await self._check_action('invalid_unix_listener', None)
 
+    @asynctest
+    async def test_rejected_tun_request(self):
+        """Test receiving inbound TUN request"""
+
+        await self._check_action('rejected_tun_request', 1)
+
+    @asynctest
+    async def test_rejected_tap_request(self):
+        """Test receiving inbound TAP request"""
+
+        await self._check_action('rejected_tap_request', 1)
+
     @asynctest
     async def test_agent_forwarding_failure(self):
         """Test failure of SSH agent forwarding"""
@@ -1464,6 +1512,15 @@ class _TestChannel(ServerTestCase):
 
             await chan.wait_closed()
 
+    @asynctest
+    async def test_ext_data_past_window(self):
+        """Test receiving an extended data packet past the advertised window"""
+
+        async with self.connect() as conn:
+            chan, _ = await _create_session(conn, 'ext_data_past_window')
+
+            await chan.wait_closed()
+
     @asynctest
     async def test_data_after_eof(self):
         """Test receiving data after EOF"""
@@ -1633,3 +1690,53 @@ class _TestChannelNoAgentForwarding(ServerTestCase):
             result = await conn.run('agent')
 
         self.assertEqual(result.exit_status, 1)
+
+
+class _TestConnectionDropbearClient(ServerTestCase):
+    """Unit tests for testing Dropbear client compatibility fix"""
+
+    @classmethod
+    async def start_server(cls):
+        """Start an SSH server to connect to"""
+
+        return await cls.create_server(_ChannelServer)
+
+    @asynctest
+    async def test_dropbear_client(self):
+        """Test reduced dropbear send packet size"""
+
+        with patch('asyncssh.connection.SSHServerChannel', _ServerChannel):
+            async with self.connect(client_version='dropbear',
+                                    max_pktsize=32759) as conn:
+                _, stdout, _ = await conn.open_session('send_pktsize')
+                self.assertEqual((await stdout.read()), '32758')
+
+            async with self.connect(client_version='dropbear',
+                                    max_pktsize=32759,
+                                    compression_algs=None) as conn:
+                _, stdout, _ = await conn.open_session('send_pktsize')
+                self.assertEqual((await stdout.read()), '32759')
+
+
+class _TestConnectionDropbearServer(ServerTestCase):
+    """Unit tests for testing Dropbear server compatibility fix"""
+
+    @classmethod
+    async def start_server(cls):
+        """Start an SSH server to connect to"""
+
+        return await cls.create_server(
+            _ChannelServer, server_version='dropbear', max_pktsize=32759)
+
+    @asynctest
+    async def test_dropbear_server(self):
+        """Test reduced dropbear send packet size"""
+
+        with patch('asyncssh.connection.SSHClientChannel', _ClientChannel):
+            async with self.connect() as conn:
+                stdin, _, _ = await conn.open_session()
+                self.assertEqual(stdin.channel.get_send_pktsize(), 32758)
+
+            async with self.connect(compression_algs=None) as conn:
+                stdin, _, _ = await conn.open_session()
+                self.assertEqual(stdin.channel.get_send_pktsize(), 32759)
