@@ -37,7 +37,7 @@ from asyncssh.socks import SOCKS5, SOCKS5_AUTH_NONE
 from asyncssh.socks import SOCKS4_OK_RESPONSE, SOCKS5_OK_RESPONSE_HDR
 
 from .server import Server, ServerTestCase
-from .util import asynctest, echo, make_certificate
+from .util import asynctest, echo, make_certificate, try_remove
 
 
 def _echo_non_async(stdin, stdout, stderr=None):
@@ -183,6 +183,18 @@ class _TCPAsyncConnectionServer(_TCPConnectionServer):
             return listen_host != 'fail'
 
 
+class _TCPAcceptHandlerServer(Server):
+    """Server for testing forwarding accept handler"""
+
+    async def server_requested(self, listen_host, listen_port):
+        """Handle a request to create a new socket listener"""
+
+        def accept_handler(_orig_host: str, _orig_port: int) -> bool:
+            return True
+
+        return accept_handler
+
+
 class _UNIXConnectionServer(Server):
     """Server for testing direct and forwarded UNIX domain connections"""
 
@@ -349,6 +361,23 @@ class _TestTCPForwarding(_CheckForwarding):
         finally:
             os.remove('.ssh/config')
 
+    @asynctest
+    async def test_proxy_jump_multiple(self):
+        """Test connecting a tunnneled SSH connection using ProxyJump"""
+
+        write_file('.ssh/config', 'Host target\n'
+                   '  Hostname localhost\n'
+                   f'  Port {self._server_port}\n'
+                   f'  ProxyJump localhost:{self._server_port},'
+                   f'localhost:{self._server_port}\n'
+                   'IdentityFile ckey\n', 'w')
+
+        try:
+            async with self.connect(host='target', username='ckey'):
+                pass
+        finally:
+            os.remove('.ssh/config')
+
     @asynctest
     async def test_proxy_jump_encrypted_key(self):
         """Test ProxyJump with encrypted client key"""
@@ -410,8 +439,10 @@ class _TestTCPForwarding(_CheckForwarding):
 
         async with self.connect() as conn:
             async with conn.listen_ssh(port=0, server_factory=Server,
-                                       server_host_keys=['skey']) as server2:
-                listen_port = server2.get_port()
+                                       server_host_keys=['skey']) as server:
+                listen_port = server.get_port()
+
+                self.assertEqual(server.get_addresses(), [('', listen_port)])
 
                 async with asyncssh.connect('127.0.0.1', listen_port,
                                             known_hosts=(['skey.pub'], [], [])):
@@ -594,6 +625,39 @@ class _TestTCPForwarding(_CheckForwarding):
                 await self._check_local_connection(listener.get_port(),
                                                    delay=0.1)
 
+    @asynctest
+    async def test_forward_local_port_accept_handler(self):
+        """Test forwarding of a local port with an accept handler"""
+
+        def accept_handler(_orig_host: str, _orig_port: int) -> bool:
+            return True
+
+        async with self.connect() as conn:
+            async with conn.forward_local_port('', 0, '', 7,
+                                               accept_handler) as listener:
+                await self._check_local_connection(listener.get_port(),
+                                                   delay=0.1)
+
+    @asynctest
+    async def test_forward_local_port_accept_handler_denial(self):
+        """Test forwarding of a local port with an accept handler denial"""
+
+        async def accept_handler(_orig_host: str, _orig_port: int) -> bool:
+            return False
+
+        async with self.connect() as conn:
+            async with conn.forward_local_port('', 0, '', 7,
+                                               accept_handler) as listener:
+                listen_port = listener.get_port()
+
+                reader, writer = await asyncio.open_connection('127.0.0.1',
+                                                               listen_port)
+
+                self.assertEqual((await reader.read()), b'')
+
+                writer.close()
+                await maybe_wait_closed(writer)
+
     @unittest.skipIf(sys.platform == 'win32',
                      'skip UNIX domain socket tests on Windows')
     @asynctest
@@ -604,7 +668,7 @@ class _TestTCPForwarding(_CheckForwarding):
             async with conn.forward_local_path_to_port('local', '', 7):
                 await self._check_local_unix_connection('local')
 
-        os.remove('local')
+        try_remove('local')
 
     @unittest.skipIf(sys.platform == 'win32',
                      'skip UNIX domain socket tests on Windows')
@@ -618,7 +682,7 @@ class _TestTCPForwarding(_CheckForwarding):
             with self.assertRaises(OSError):
                 await conn.forward_local_path_to_port('local', '', 7)
 
-        os.remove('local')
+        try_remove('local')
 
     @asynctest
     async def test_forward_local_port_pause(self):
@@ -751,7 +815,8 @@ class _TestTCPForwarding(_CheckForwarding):
 
         server.close()
         await server.wait_closed()
-        os.remove('local')
+
+        try_remove('local')
 
     @asynctest
     async def test_forward_remote_specific_port(self):
@@ -855,6 +920,33 @@ class _TestTCPForwarding(_CheckForwarding):
             await listener.wait_closed()
 
 
+class _TestTCPForwardingAcceptHandler(_CheckForwarding):
+    """Unit tests for TCP forwarding with accept handler"""
+
+    @classmethod
+    async def start_server(cls):
+        """Start an SSH server which supports TCP connection forwarding"""
+
+        return await cls.create_server(
+            _TCPAcceptHandlerServer, authorized_client_keys='authorized_keys')
+
+    @asynctest
+    async def test_forward_remote_port_accept_handler(self):
+        """Test forwarding of a remote port with accept handler"""
+
+        server = await asyncio.start_server(echo, None, 0,
+                                            family=socket.AF_INET)
+        server_port = server.sockets[0].getsockname()[1]
+
+        async with self.connect() as conn:
+            async with conn.forward_remote_port(
+                    '', 0, '127.0.0.1', server_port) as listener:
+                await self._check_local_connection(listener.get_port())
+
+        server.close()
+        await server.wait_closed()
+
+
 class _TestAsyncTCPForwarding(_TestTCPForwarding):
     """Unit tests for AsyncSSH TCP connection forwarding with async return"""
 
@@ -946,7 +1038,7 @@ class _TestUNIXForwarding(_CheckForwarding):
             await listener.wait_closed()
             listener.close()
 
-        os.remove('echo')
+        try_remove('echo')
 
     @asynctest
     async def test_unix_server_open(self):
@@ -979,7 +1071,7 @@ class _TestUNIXForwarding(_CheckForwarding):
             async with conn.start_unix_server(_unix_listener_non_async, path):
                 await self._check_local_unix_connection('echo')
 
-        os.remove('echo')
+        try_remove('echo')
 
     @asynctest
     async def test_unix_server_failure(self):
@@ -997,7 +1089,40 @@ class _TestUNIXForwarding(_CheckForwarding):
             async with conn.forward_local_path('local', '/echo'):
                 await self._check_local_unix_connection('local')
 
-        os.remove('local')
+        try_remove('local')
+
+    @asynctest
+    async def test_forward_local_port_to_path_accept_handler(self):
+        """Test forwarding of port to UNIX path with accept handler"""
+
+        def accept_handler(_orig_host: str, _orig_port: int) -> bool:
+            return True
+
+        async with self.connect() as conn:
+            async with conn.forward_local_port_to_path(
+                    '', 0, '/echo', accept_handler) as listener:
+                await self._check_local_connection(listener.get_port(),
+                                                   delay=0.1)
+
+    @asynctest
+    async def test_forward_local_port_to_path_accept_handler_denial(self):
+        """Test forwarding of port to UNIX path with accept handler denial"""
+
+        async def accept_handler(_orig_host: str, _orig_port: int) -> bool:
+            return False
+
+        async with self.connect() as conn:
+            async with conn.forward_local_port_to_path(
+                    '', 0, '/echo', accept_handler) as listener:
+                listen_port = listener.get_port()
+
+                reader, writer = await asyncio.open_connection('127.0.0.1',
+                                                               listen_port)
+
+                self.assertEqual((await reader.read()), b'')
+
+                writer.close()
+                await maybe_wait_closed(writer)
 
     @asynctest
     async def test_forward_local_port_to_path(self):
@@ -1042,8 +1167,8 @@ class _TestUNIXForwarding(_CheckForwarding):
         server.close()
         await server.wait_closed()
 
-        os.remove('echo')
-        os.remove('local')
+        try_remove('echo')
+        try_remove('local')
 
     @asynctest
     async def test_forward_remote_path_to_port(self):
@@ -1060,11 +1185,11 @@ class _TestUNIXForwarding(_CheckForwarding):
                     path, '127.0.0.1', server_port):
                 await self._check_local_unix_connection('echo')
 
-        os.remove('echo')
-
         server.close()
         await server.wait_closed()
 
+        try_remove('echo')
+
     @asynctest
     async def test_forward_remote_path_failure(self):
         """Test failure of forwarding a remote UNIX domain path"""
@@ -1077,7 +1202,7 @@ class _TestUNIXForwarding(_CheckForwarding):
             with self.assertRaises(asyncssh.ChannelListenError):
                 await conn.forward_remote_path(path, 'local')
 
-        os.remove('echo')
+        try_remove('echo')
 
     @asynctest
     async def test_forward_remote_path_not_permitted(self):
