@@ -24,6 +24,7 @@ import asyncio
 import binascii
 import functools
 import os
+import shutil
 import subprocess
 import sys
 import tempfile
@@ -31,9 +32,15 @@ import unittest
 
 from unittest.mock import patch
 
-from cryptography.hazmat.backends.openssl import backend
+from asyncssh import set_default_skip_rsa_key_validation
+from asyncssh.gss import gss_available
+from asyncssh.logging import logger
+from asyncssh.misc import ConnectionLost, SignalReceived
+from asyncssh.packet import Byte, String, UInt32, UInt64
+from asyncssh.public_key import generate_private_key
 
-# pylint: disable=unused-import
+
+# pylint: disable=ungrouped-imports, unused-import
 
 try:
     import bcrypt
@@ -41,6 +48,8 @@ try:
 except ImportError: # pragma: no cover
     bcrypt_available = False
 
+nc_available = bool(shutil.which('nc'))
+
 try:
     import uvloop
     uvloop_available = True
@@ -53,14 +62,7 @@ try:
 except ImportError: # pragma: no cover
     x509_available = False
 
-# pylint: enable=unused-import
-
-from asyncssh.gss import gss_available
-from asyncssh.logging import logger
-from asyncssh.misc import ConnectionLost, SignalReceived
-from asyncssh.packet import Byte, String, UInt32, UInt64
-from asyncssh.public_key import generate_private_key
-
+# pylint: enable=ungrouped-imports, unused-import
 
 # pylint: disable=no-member
 
@@ -74,15 +76,10 @@ else: # pragma: no cover
 # pylint: enable=no-member
 
 
-# pylint: disable=protected-access
-
-# Disable RSA key blinding to speed up unit tests
-backend._rsa_skip_check_key = True
-
-# pylint: enable=protected-access
-
 _test_keys = {}
 
+set_default_skip_rsa_key_validation(True)
+
 
 def asynctest(coro):
     """Decorator for async tests, for use with AsyncTestCase"""
@@ -109,6 +106,20 @@ def patch_getnameinfo(cls):
     return patch('socket.getnameinfo', getnameinfo)(cls)
 
 
+def patch_extra_kex(cls):
+    """Decorator for skipping extra kex algs"""
+
+    def skip_extra_kex_algs(self):
+        """Don't send extra key exchange algorithms"""
+
+        # pylint: disable=unused-argument
+
+        return []
+
+    return patch('asyncssh.connection.SSHConnection._get_extra_kex_algs',
+                 skip_extra_kex_algs)(cls)
+
+
 def patch_gss(cls):
     """Decorator for patching GSSAPI classes"""
 
@@ -224,6 +235,15 @@ def run(cmd):
         raise
 
 
+def try_remove(filename):
+    """Try to remove a file, ignoring errors"""
+
+    try:
+        os.remove(filename)
+    except OSError: # pragma: no cover
+        pass
+
+
 class ConnectionStub:
     """Stub class used to replace an SSHConnection object"""
 
